"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const opentracing_1 = require("opentracing");
const util_1 = require("util");
const base_1 = require("./base");
const util_2 = require("./util");
class Fork extends base_1.default {
    constructor() {
        super(...arguments);
        // Bifurcaciones
        this.Forks = [];
        // Bandera de la primera Bifurcacion
        this.isFirstFork = false;
        this.configForks = new Map();
    }
    /**
     * Metodo que hace rel registro de Forks
     * para su utilizacion sobre los servicios.
     * @param conf Configuracion para el evento
     */
    forksConfig(conf) {
        this.Forks = [];
        conf.forEach((element) => {
            // Id unico
            const id = util_2.default.key();
            // Almacenar
            if (element.conf.App === 'appcoppel' &&
                element.conf.Version === 'v1') {
                this.configForks.set(element.conf.Name, element);
            }
            else {
                this.configForks.set(`${element.conf.App}_${element.conf.Version}_${element.conf.Name}`, element);
            }
            // Agregar el id
            element.conf.id = id;
            // Agregar a la coleccion
            this.Forks.push(element.conf);
        });
    }
    /**
     * Metodo que procesa los mensajes
     * @param data Datos
     * @param Method Metodo que se ejecutara
     */
    message(data, Method, params = []) {
        const log = {
            event: 'input',
        };
        params.forEach((value) => {
            if (value === 'uuid') {
                log.uuid = data.metadata.uuid;
            }
            else if (value === 'authorizacion') {
                log.authorizacion = data.headers.Authorizacion;
            }
            else if (value === 'data') {
                log.data = data.data;
            }
            else if (value === 'response') {
                if (data.response &&
                    data.response.data &&
                    data.response.data.response) {
                    log.response = data.response.data.response;
                }
                else {
                    log.response = data.data;
                }
            }
        });
        this.span.log(log);
        // Posicion del actual Fork
        let nextFork = this.nextFork(data.metadata.owner);
        // Agregar el tiempo actual
        data.metadata.time = new Date().toISOString();
        // Indicar tipo
        data.metadata.mtype = 'input';
        // Retornar la promesa
        return new Promise((rr, ff) => {
            // Ver si ya es una bifurcacion.
            if (this.Topic !== data.metadata.owner || this.Forks.length < 1) {
                // Recuperar el siguiente
                nextFork = this.nextFork(data.metadata.owner);
                // Validar si ocurrio un error
                const isError = this.isError(data);
                // Validar si aun se tienen elementos que procesar (Forks)
                if (util_1.isNull(nextFork) || isError) {
                    // Indicar que ya no es bifurcacion
                    data.metadata.bifurcacion = false;
                }
                else {
                    data.metadata.bifurcacion = true;
                }
                // Ver si ocurrio un error
                if (isError) {
                    // -.-
                    const r = this.formatResponse(data, null, this.getErrorCode(data), true, this.getErrorMeta(data));
                    // Regresar la respuesta
                    ff(r);
                }
                else if (util_1.isNull(nextFork) || util_1.isUndefined(nextFork)) {
                    // Ya no existen mas paso, ejecutar la funcion principal.
                    try {
                        const paraInject = [];
                        params.forEach((value) => {
                            if (value === 'uuid') {
                                paraInject.push(data.metadata.uuid);
                            }
                            else if (value === 'filters') {
                                paraInject.push(this.transformFilters(data.metadata.filters));
                            }
                            else if (value === 'authorizacion') {
                                paraInject.push(data.headers.Authorizacion);
                            }
                            else if (value === 'data') {
                                paraInject.push(data.metadata.data_inicial);
                            }
                            else if (value === 'response') {
                                if (data.response &&
                                    data.response.data &&
                                    data.response.data.response) {
                                    paraInject.push(data.response.data.response);
                                }
                                else {
                                    paraInject.push(data.data);
                                }
                            }
                            else if (value === 'parentSpan') {
                                paraInject.push(this.span);
                            }
                        });
                        const resp = Method.apply({}, paraInject);
                        // Validar si la respuesta es una promesa y esperar a resolverla
                        if (util_2.default.isPromise(resp)) {
                            // Esperar que finalice
                            resp.then((response) => {
                                // Consultar el estado
                                const format = this.httpFormat(data, response);
                                // Crear la salida
                                const r = this.formatResponse(format.request, format.data, 0);
                                // Regresar el contenido
                                rr(r);
                            }).catch((error) => {
                                this.span.setTag(opentracing_1.Tags.ERROR, true);
                                this.span.log({
                                    'error.message': error.message,
                                    'error.object': error,
                                    'error.stack': error.stack,
                                    // tslint:disable-next-line: object-literal-key-quotes
                                    event: 'error',
                                });
                                // Indicar que fue error
                                this.errorHandler(data, error, ff);
                            });
                        }
                        else {
                            // Consultar el estado
                            const format = this.httpFormat(data, resp);
                            // Crear la salida
                            const r = this.formatResponse(format.request, format.data, 0);
                            // Regresar el contenido
                            rr(r);
                        }
                    }
                    catch (error) {
                        this.span.setTag(opentracing_1.Tags.ERROR, true);
                        this.span.log({
                            'error.message': error.message,
                            'error.object': error,
                            'error.stack': error.stack,
                            // tslint:disable-next-line: object-literal-key-quotes
                            event: 'error',
                        });
                        // Regresar el error
                        this.errorHandler(data, error, ff);
                    }
                }
                else {
                    // Recuperar los datos requeridos del siguiente Topico
                    const row = this.getNextTopic(nextFork);
                    // Recuperar la configuracion
                    const conf = this.configForks.get(row.current);
                    // Agregar el metodo que corresponde y se guarda una copia para usarla despues
                    data.metadata.original_method = data.metadata.method;
                    data.metadata.method = conf.action;
                    let resp = null;
                    if (data.response &&
                        data.response.data &&
                        data.response.data.response) {
                        resp = data.response.data.response;
                    }
                    // Recuperar la configuracion
                    try {
                        data.data = this.getData(data.metadata.data_inicial, conf, resp);
                    }
                    catch (error) {
                        // Enviar el error
                        this.errorHandler(data, error, ff);
                        return;
                    }
                    // Aplicar los cambios
                    data = this.changeMetadata(data, row.current);
                    // Consultar el estado
                    const format = this.httpFormat(data, null);
                    const r = this.formatResponse(format.request, format.data, 0);
                    // const r: any = this.formatResponse(data, null, 0);
                    // Enviar la peticion para su procesamiento
                    rr(r);
                }
            }
            else {
                // Recuperar los datos requeridos del siguiente Topico
                const row = this.getNextTopic(nextFork);
                // Recuperar la configuracion
                const conf = this.configForks.get(row.current);
                // Indicar que es una bifurcacion
                data.metadata.bifurcacion = true;
                // Agregar el metodo que corresponde y se guarda una copia para usarla despues
                data.metadata.original_method = data.metadata.method;
                data.metadata.method = conf.action;
                // Recuperar la configuracion
                try {
                    data.data = this.getData(data.data, conf, null);
                }
                catch (error) {
                    // Enviar el error
                    this.errorHandler(data, error, ff);
                    return;
                }
                // Aplicar los cambios
                data = this.changeMetadata(data, row.current);
                // Consultar el estado
                const format = this.httpFormat(data, null);
                const r = this.formatResponse(format.request, format.data, 0);
                // const r: any = this.formatResponse(data, null, 0);
                // Enviar la peticion para su procesamiento
                rr(r);
            }
        });
    }
    /**
     * Metodo para recuperar el data que se ejecutara
     * @param dataDefault Data por default
     * @param conf Configuracion del fork
     * @param response Datos del servicio previo
     */
    getData(dataDefault, conf, response) {
        if (conf.fnc) {
            // Ejecutar la funcion
            const resp = conf.fnc.apply({}, [dataDefault, response]);
            // Validar que sea un objeto
            if (typeof resp === 'object') {
                return resp;
            }
            else {
                return { value: resp };
            }
        }
        else if (conf.data) {
            return conf.data;
        }
        else {
            // No coincide nada regresar datos por default.
            return dataDefault;
        }
    }
    /**
     * Metodo para crear la estructura para la publicacion
     * del mensaje.
     * @param worker Configuracion del worker
     */
    getNextTopic(worker) {
        let current = `${worker.App}_${worker.Version}_${worker.Name}`;
        if (worker.Version === 'v1' && worker.App === 'appcoppel') {
            current = worker.Name;
        }
        return { current, name: worker.Name };
    }
    /**
     * Metodo para buscar la posicion del siguiente
     * Fork a ejecutar
     * @param nameFork Nombre que se busca
     */
    nextFork(nameFork) {
        // Ver el tamaño las bifurcaciones asignadas
        const size = this.Forks.length;
        // Ver si no corresponde al propio topico
        if (nameFork === this.Topic) {
            // Indicar que es la primera iteracion
            this.isFirstFork = true;
            // Regresar el primer elemento.
            return this.Forks[0];
        }
        // Buscar la posicion del elemento actual
        for (let index = 0; index < size; index++) {
            const obj = this.Forks[index];
            const { current, name } = this.getNextTopic(obj);
            // Validar si es el que corresponde
            if (current === nameFork) {
                if (index + 1 < size) {
                    // Regresar el siguiente Fork
                    return this.Forks[index + 1];
                }
                else {
                    return null;
                }
            }
        }
        // Indicar que es la primera iteracion
        this.isFirstFork = true;
        // Regresar el primer elemento.
        return this.Forks[0];
    }
    /**
     * Metodo para asergurar si ocurrio un error al momento de consumir el servicio
     * @param data Datos a procesar
     */
    isError(data) {
        // Realizar la validacion
        return (data.response &&
            data.response.data.response &&
            !util_1.isUndefined(data.response.data.response.errorCode));
    }
    /**
     * Metodo para recuperar el error code
     * @param data Data a procesar
     */
    getErrorCode(data) {
        return data.response.data.response.errorCode || -99;
    }
    /**
     * Metodo para crear el error Meta que se pasa como mensaje de error.
     *
     * @private
     * @param {DataRequest} data
     * @returns {ErrorMetadata}
     * @memberof Fork
     */
    getErrorMeta(data) {
        // Datos del servicio
        let service;
        // Recuperar el topico de donde viene el error.
        const topic = data.metadata.owner;
        // Crear las piezas segun la separacion de guion bajo
        const parts = topic.split('_');
        // La version legada cuenta con un solo elemento, las versiones nuevas 3.
        if (parts.length < 2) {
            // Solo es un elemento, llenar con los datos por default
            service = {
                App: 'coppelapp',
                Version: 'v1',
                Service: topic,
            };
        }
        else {
            // Version nueva.
            service = {
                App: parts[0],
                Version: parts[1] || 'v1',
                Service: parts[2] || 'No identificado',
            };
        }
        // Regresar el error correcto
        return {
            errorCode: data.response.data.response.errorCode || -99,
            userMessage: data.response.data.response.userMessage ||
                'No se proporciono un mensaje de error, desde el servicio.',
            Service: Object.assign(service, {
                Framework: data.response.data.response.framework || 'No especificado',
            }),
        };
    }
    /**
     * Metodo apra aplicar los cambios sobre
     * la metadata
     * @param data Data a cambiar
     * @param current Worker Actual
     */
    changeMetadata(data, current) {
        const operationId = new Date().getTime() / 1000000;
        // Aplicar los cambios requeridos
        data.metadata.id_operacion = Math.floor(operationId);
        data.metadata.uowner = data.metadata.uworker; // original
        data.metadata.worker = current; // Worker que procesara
        data.metadata.uworker = `${data.metadata.worker}_${data.metadata.id_operacion}`;
        data.metadata.owner = this.Topic; // callback
        // Regresar los nuevos datos
        return data;
    }
}
exports.default = Fork;
