"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const opentracing_1 = require("opentracing");
const util_1 = require("./util");
const Fork_1 = require("./Fork");
const BPromise = require("bluebird");
const axios_1 = require("axios");
const util_2 = require("util");
const Error_1 = require("./Error");
class ForkHttp extends Fork_1.default {
    constructor(conf, topic, span, tracer) {
        super(conf, topic, span);
        this.services = [];
        this.tracer = tracer;
    }
    forksConfig(conf) {
        this.services = conf;
    }
    message(data, Method, params) {
        // Iniciar el proceso
        return new Promise((rr, ff) => __awaiter(this, void 0, void 0, function* () {
            // rr(this.formatResponse(data, null, 0));
            const paraInject = [];
            // Validar que exista la direccion a
            if (util_2.isUndefined(process.env.URL_ENDPOINT_SERVICES)) {
                ff(this.formatResponse(data, null, -96));
            }
            else {
                // Respuesta anterior
                let dataResponseService = null;
                // Ejecutar todas las peticions a los servicios REST.
                BPromise.each(this.services, (service) => __awaiter(this, void 0, void 0, function* () {
                    // ========== INICIA LA PETICION ==============
                    const spanOperation = this.tracer.startSpan(`${service.conf.App}_${service.conf.Version}_${service.conf.Name}`, {
                        childOf: this.span.context(),
                    });
                    // Url del servicio
                    const url = process.env.URL_ENDPOINT_SERVICES;
                    // Generar la URL del servicio
                    const urlservice = `${url}/${service.conf.App}/api/${service.conf.Version}/${service.conf.Name}`;
                    // Respuesta de la ejecucion de la funcion asignada
                    let respFnc = null;
                    if (util_2.isFunction(service.fnc)) {
                        // Ejecutar al funcion
                        try {
                            // Ejecutamos la funcion y recuperamos su contenido
                            respFnc = yield this.getResponse(service.fnc.apply({}, [
                                data.data,
                                dataResponseService,
                            ]));
                        }
                        catch (error) {
                            //
                            ff(this.formatResponse(data, null, -99));
                            spanOperation.setTag(opentracing_1.Tags.ERROR, true);
                            spanOperation.finish();
                            return;
                        }
                    }
                    // Retornar la promesa de la peticion.
                    return new Promise((resolve, reject) => {
                        // Asingar la peticion
                        spanOperation.setTag(opentracing_1.Tags.HTTP_METHOD, service.action);
                        // Indicar la URL
                        spanOperation.setTag(opentracing_1.Tags.HTTP_URL, urlservice);
                        // Indicar que operacion se aplico
                        spanOperation.log({
                            type: 'request',
                            data: respFnc !== null && respFnc !== void 0 ? respFnc : service.data,
                        });
                        // Aplicar la peticion
                        axios_1.default({
                            method: service.action,
                            url: urlservice,
                            data: respFnc !== null && respFnc !== void 0 ? respFnc : service.data,
                        })
                            .then((rrp) => {
                            // ========== FINALIZA LA PETICION ===========
                            // Validar el formato de la salida
                            if (rrp.data.hasOwnProperty('meta') &&
                                rrp.data.hasOwnProperty('data')) {
                                if (rrp.data.data.hasOwnProperty('response')) {
                                    // Almacenar la respuesta
                                    dataResponseService =
                                        rrp.data.data.response;
                                    // Verificar si no es un mensaje de error.
                                    if (dataResponseService.hasOwnProperty('errorCode')) {
                                        //
                                        spanOperation.setTag(opentracing_1.Tags.ERROR, true);
                                        spanOperation.log(rrp.data);
                                        spanOperation.finish();
                                        // Retornar el error
                                        this.errorHandler(data, new Error_1.MsError(dataResponseService.userMessage, dataResponseService.errorCode, 500), ff);
                                        reject(null);
                                    }
                                    else {
                                        // Finalizar la promesa
                                        spanOperation.log(rrp.data);
                                        spanOperation.finish();
                                        resolve(rrp.data);
                                    }
                                }
                                else {
                                    spanOperation.setTag(opentracing_1.Tags.ERROR, true);
                                    spanOperation.finish();
                                    this.errorHandler(data, new Error_1.MsError(`El servicio ${service.conf.Name} retorno una respuesta con un formato desconocido`, -99, 400), ff);
                                    reject(null);
                                    return;
                                }
                            }
                            else {
                                spanOperation.setTag(opentracing_1.Tags.ERROR, true);
                                spanOperation.finish();
                                this.errorHandler(data, new Error_1.MsError(`El servicio ${service.conf.Name} retorno una respuesta con un formato desconocido`, -99, 400), ff);
                                reject(null);
                                return;
                            }
                        })
                            .catch((reason) => {
                            // ========== FINALIZA LA PETICION ===========
                            spanOperation.setTag(opentracing_1.Tags.ERROR, true);
                            // Error
                            let ErrorRequest = null;
                            // Recuperamos la respuesta
                            const respReq = reason.response.data;
                            // Validar si es una respuesta del servicio
                            if (respReq.hasOwnProperty('meta') &&
                                respReq.hasOwnProperty('data')) {
                                if (respReq.data.hasOwnProperty('response')) {
                                    const response = respReq.data.response;
                                    // Crear el cuerpo del error
                                    ErrorRequest = new Error_1.MsError(response.userMessage, response.errorCode, reason.response.status);
                                }
                                else {
                                    // Lanzar error generico
                                    // Crear el cuerpo del error
                                    ErrorRequest = new Error_1.MsError(`Ocurrio una excepción al consumir el servicio ${service.conf.Name}`, -99, reason.response.status);
                                }
                            }
                            else {
                                // Lanzar error generico
                                // Crear el cuerpo del error
                                ErrorRequest = new Error_1.MsError(`Ocurrio una excepción al consumir el servicio ${service.conf.Name}`, -99, reason.response.status
                                    ? reason.response
                                    : 500);
                            }
                            // Etiquetas
                            if (reason.hasOwnProperty('response')) {
                                if (reason.response.hasOwnProperty('status') &&
                                    reason.response.hasOwnProperty('data')) {
                                    spanOperation.setTag(opentracing_1.Tags.HTTP_STATUS_CODE, reason.response.status);
                                    spanOperation.log(reason.response.data);
                                }
                            }
                            // Finalizar el span
                            spanOperation.finish();
                            // Dar formato al error
                            this.errorHandler(data, ErrorRequest, ff);
                            reject(null);
                            return;
                        });
                    });
                }))
                    .then((result) => {
                    try {
                        params.forEach((value) => {
                            if (value === 'uuid') {
                                paraInject.push(data.metadata.uuid);
                            }
                            else if (value === 'filters') {
                                paraInject.push(this.transformFilters(data.metadata.filters));
                            }
                            else if (value === 'authorizacion') {
                                paraInject.push(data.headers.Authorizacion);
                            }
                            else if (value === 'data') {
                                paraInject.push(data.data);
                            }
                            else if (value === 'response') {
                                paraInject.push(dataResponseService);
                            }
                            else if (value === 'parentSpan') {
                                paraInject.push(this.span);
                            }
                        });
                        const resp = Method.apply({}, paraInject);
                        // Validar si la respuesta es una promesa y esperar a resolverla
                        if (util_1.default.isPromise(resp)) {
                            // Esperar que finalice
                            resp.then((response) => {
                                // Regresar el contenido
                                rr(this.formatResponse(data, response, 0));
                            }).catch((error) => {
                                this.span.setTag(opentracing_1.Tags.ERROR, true);
                                this.span.log({
                                    'error.message': error.message,
                                    'error.object': error,
                                    'error.stack': error.stack,
                                    // tslint:disable-next-line: object-literal-key-quotes
                                    event: 'error',
                                });
                                // Indicar que fue error
                                ff(this.formatResponse(data, null, error.code));
                            });
                        }
                        else {
                            // Regresar la respuesta
                            rr(this.formatResponse(data, resp, 0));
                        }
                    }
                    catch (error) {
                        this.span.setTag(opentracing_1.Tags.ERROR, true);
                        this.span.log({
                            'error.message': error.message,
                            'error.object': error,
                            'error.stack': error.stack,
                            // tslint:disable-next-line: object-literal-key-quotes
                            event: 'error',
                        });
                        // Regresar el error
                        ff(this.formatResponse(data, null, error.code));
                    }
                })
                    .catch((rejection) => {
                    //
                });
            }
        }));
    }
    /**
     * Metodo para recuperar el resultado de un metodo.
     *
     * @param {*} data
     * @returns {*}
     * @memberof ForkHttp
     */
    getResponse(data) {
        return new Promise((rr, ff) => {
            if (util_1.default.isPromise(data)) {
                data.then((rc) => {
                    rr(rc);
                }).then((e) => {
                    ff(e);
                });
            }
            else {
                rr(data);
            }
        });
    }
}
exports.default = ForkHttp;
